/*
 * oggThumb creates thumbs from an ogg/theora video
 *
 * Copyright (C) 2008-2009 Joern Seger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef __WIN32
#define __GNU_LIBRARY__
#include "../win32/getopt_win.h"
#endif

#include <iostream>
#include <map>
#include <vector>
#include <string>
#include <sstream>

#include <cstring>
#include <cstdlib>
#include <cmath>
#include <ctime>

#include "definition.h"
#include "theoraEncoder.h"
#include "theoraDecoder.h"
#include "streamSerializer.h"
#include "streamMux.h"
#include "videoFormats.h"
#include "theoraStreamParameter.h"
#include "fileRepository.h"

const std::string validChars("0123456789,x");

void printHelpScreen(const std::string& progname)
{
	std::cerr << "usage: "<<progname<<" [Options] originalFile.ogm newFile.ogm\n"
	          << "  Option: \n"
	          << "    -h     this helpscreen\n"
	          << "    -s     <width x height>  new frame size\n"
	          << "    -f     new framerate\n"
	          << "    -d     datarate in kBit/s\n";
//              << "    -t     stretch picture to new size\n";

}

void extractUint32(std::deque<uint32>& list, const std::string& _argument,
    char seperator)
{
  std::string argument(_argument);
  std::stringstream str;
  std::string substr;

  // if there is no argument given, the first frame will be created as a thumbnail
  if (argument.empty()) {
    list.push_back(0);
    return;
  }

  // delete all invalid data
  std::size_t pos;
  while ((pos = argument.find_first_not_of(validChars)) != std::string::npos){
    std::cerr << "erasing <"<<argument.at(pos)<<">\n";
    argument.erase(pos,1);
  }

  str << argument;

  uint32 value(0);
  while (!str.eof()) {
    std::stringstream part;
    getline(str, substr, seperator);
    part << substr;
    part >> value;
    list.push_back(value);
  }

}

int main(int argc, char* argv[])
{
	  uint32 width(480);
	  uint32 height(320);
	  uint32 framesPerSecond(24);
	  uint32 datarate(256000);
	  //float  resample(1.2);
	  bool strech(false);

	  TheoraStreamParameter configOutput;
	  TheoraStreamParameter configInput;

	  std::string programName(argv[0]);

	  srand(time(0));

	  int opt;
	  while ((opt = getopt(argc, argv, "hs:f:d:t:")) != EOF)

	    switch (opt) {

	    case 'h':
	    case '?':
	      printHelpScreen(programName);
	      exit(-1);

	    case 's': {
	      std::deque<uint32> framesize;
	      extractUint32(framesize, optarg, 'x');
	      if (framesize.size() != 2) {
	        std::cerr << "please specify the size in the following way: -p320x480\n";
	        exit(-1);
	      }
	      width = framesize[0];
	      height = framesize[1];

	    } break;

	    case 'f':
	      framesPerSecond = atoi(optarg);
	      break;

	    case 'd':
	      datarate = atoi(optarg); // yes, I know the atoi bug
	      break;

	    case 't':
	    	strech = true;
	      break;


	    }

	  argc -= optind;
	  argv += optind;

	  if ((argc < 2)) {
	    std::cerr << "usage: oggResize [options] <inputfile> <outputfile>\n";
	    return (-1);
	  }

	  std::string inputFile  = std::string(argv[0]);
	  std::string outputFile = std::string(argv[1]);


	  /* create configuration */

	  StreamSerializer inStream;
	  if (!inStream.open(inputFile)) {
		std::cerr << "can not open file <"<<inputFile<<">\n";
		return(-1);
	  }

	  /* get all information from the stream */
	  std::vector<StreamConfig> inFileConfigList;
	  std::vector<StreamConfig> outFileConfigList;

	  inStream.getStreamConfig(inFileConfigList);

	  bool foundTheora(false);
	  int8 theoraStreamID(-1);

	  /* Output some stream information */
	  for (uint32 i(0); i<inFileConfigList.size(); ++i) {
	    StreamConfig& conf(inFileConfigList[i]);
	    std::cerr << "Stream No: "<<(int)conf.streamNo<<"(0x"<<std::hex
	        << conf.serialNo<<std::dec<<")\n"
	        << conf.parameter->toString();
	    if (conf.type == ogg_theora)  {
	      foundTheora = true;
	      theoraStreamID = i;

	      configInput = *(TheoraStreamParameter*)(conf.parameter);
	    }
	  }

	  if (foundTheora == false) {
		std::cerr << "Does not find any theora stream within file <"<<inputFile<<">\n";
		return(-1);
	  }

	  // copy all config information
	  // the theora stream is changed after that
	  outFileConfigList = inFileConfigList;

	  uint32 widthIn = configInput.pictureX;

	  configOutput.videoBitrate = datarate;
	  configOutput.pictureX = width;
	  configOutput.pictureY = height;
//	  configOutput.frameX   = width;
//	  configOutput.frameY   = height;
//	  configOutput.frameXOffset = ;
//	  configOutput.frameYOffset = 0;
	  configOutput.aspectRatioDenom = 1;
	  configOutput.aspectRatioNum = 1;
	  configOutput.framerateNum = framesPerSecond;
	  configOutput.framerateDenom = 1;
	  configOutput.keyframeShift = 6;


	  TheoraDecoder theoraDecoder;
	  theoraDecoder.initDecoder(inFileConfigList[theoraStreamID]);

	  /* create stream configuration */
	  TheoraEncoder theoraEncoder(0);

	  // this configuration is filled by the encoder
	  StreamConfig newTheoraStreamConf;

//	  std::cerr << "Output befor:\n" << configOutput.toString();

	  /* configure the theora encoder and get a stream config back
	   * which configures the stream multiplexer */
	  try {
	  theoraEncoder.configureEncoder(configOutput, newTheoraStreamConf);
	  } catch(const char* data) {
	    std::cerr << data;
	    exit(-1);
	  }

//	  std::cerr << "Output after:\n" << configOutput.toString();

	  // encoder might want another frame size:
	  width = configOutput.frameX;
	  height = configOutput.frameY;

	  outFileConfigList[theoraStreamID] = newTheoraStreamConf;

	  /* create a repository, where the data should be placed */
	  FileRepository repository(outputFile, MediaUnit::write);

	  /* create a stream multiplexer */
	  StreamMux streamCreate(&repository);

	  /* configure the stream multiplexer */
	  streamCreate.configureStreams(outFileConfigList);

	  OggPacket packet;
	  OggPacket newPacket;
	  double time;
	  double distance((1.0*configOutput.framerateDenom)/(1.0*configOutput.framerateNum));
	  double nextTime(0);
	  th_ycbcr_buffer inycbcr;
	  th_ycbcr_buffer outycbcr;
	  //YCbCrPlane inPlane;
	  //YCbCrPlane outPlane;

	  RGBPlane inPlane;
	  RGBPlane outPlane;

#ifdef HAVE_BZERO
	  bzero(&outycbcr,sizeof(th_ycbcr_buffer));
#else
          for(uint32 i(0);i<sizeof(th_ycbcr_buffer);++i) *(((uint8*)&outycbcr)+i) = 0;
#endif

	  uint8 streamNo((uint8)outFileConfigList[theoraStreamID].streamNo);

//	  std::cerr << theoraDecoder.getInfoString();

	  while(inStream.available()) {
		  time = inStream.getNextPacket(packet);

		  if (packet.getStreamNo() == streamNo) {
			// relevant packet
			  try {
				  theoraDecoder << packet;
			  theoraDecoder >> inycbcr;

			  std::cerr << "\r  "<<time<<"              ";//<< inycbcr[0].width << " "<<inycbcr[0].stride<<" "<<widthIn<<"           \r";

			  /* for some reason, the stride is not correct in some cases */
			  inycbcr[0].width = widthIn;
			  inycbcr[1].width = widthIn/2;
			  inycbcr[2].width = widthIn/2;

			  while (time > nextTime) {
			  inPlane.cleanup();
			  inPlane.importYCrCb_theora(inycbcr);

			  outPlane.cleanup();
/*			  if (stretch)
				  outPlane = inPlane.resize(width, height);
			  else
*/
			  outPlane = inPlane.reframe(width, height);

			  outPlane.exportYCrCb_theora(outycbcr);

			  theoraEncoder << outycbcr;
			  theoraEncoder >> packet;

			  streamCreate << packet;
			  nextTime += distance;
			  //std::cerr <<" "<< nextTime <<" ";
			  }
			  }
			  catch (char* error) {
				  std::cerr << error;
			  }
		  }
		  else {
			// other packet
			streamCreate << packet;

		  }

	  }

	  streamCreate.setEndOfStream();
	  std::cerr << std::endl;
}